"""
The parameters defined in the FIPS 203 document.

Includes the ML-KEM-512, ML-KEM-768, and ML-KEM-1024 parameters
and initialised objects with them.
"""

from .ml_kem import ML_KEM

# TODO: we can only allow a user to select one of the following three
# we should maybe put these into the class and only allow a user to
# select 128, 192 or 256 bit security.
DEFAULT_PARAMETERS = {
    "ML512": {
        "k": 2,
        "eta_1": 3,
        "eta_2": 2,
        "du": 10,
        "dv": 4,
        "oid": (2, 16, 840, 1, 101, 3, 4, 4, 1),
    },
    "ML768": {
        "k": 3,
        "eta_1": 2,
        "eta_2": 2,
        "du": 10,
        "dv": 4,
        "oid": (2, 16, 840, 1, 101, 3, 4, 4, 2),
    },
    "ML1024": {
        "k": 4,
        "eta_1": 2,
        "eta_2": 2,
        "du": 11,
        "dv": 5,
        "oid": (2, 16, 840, 1, 101, 3, 4, 4, 3),
    },
}
"""Parameters for the :py:obj:`.ML_KEM` objects."""

ML_KEM_512 = ML_KEM(DEFAULT_PARAMETERS["ML512"])
"""
Key exchange object that uses ML-KEM-512 parameters internally.

Provides about 128 bit level of security.

Part of stable API.
"""

ML_KEM_768 = ML_KEM(DEFAULT_PARAMETERS["ML768"])
"""
Key exchange object that uses ML-KEM-768 parameters internally.

Provides about 192 bit level of security.

Part of stable API.
"""

ML_KEM_1024 = ML_KEM(DEFAULT_PARAMETERS["ML1024"])
"""
Key exchange object that uses ML-KEM-1024 parameters internally.

Provides about 256 bit level of security.

Part of stable API.
"""



import random


class GenericPolynomialRing:
    """
    Initialise the polynomial ring:

        R = GF(q) / (X^n + 1)
    """

    def __init__(self, q, n):
        self.q = q
        self.n = n
        self.element = GenericPolynomial

    def zero(self):
        """
        Return the zero element of the polynomial ring
        """
        return self([0])

    def gen(self):
        """
        Return the generator `x` of the polynomial ring
        """
        return self([0, 1])

    def random_element(self):
        """
        Compute a random element of the polynomial ring with coefficients in the
        canonical range: ``[0, q-1]``
        """
        coefficients = [random.randint(0, self.q - 1) for _ in range(self.n)]
        return self(coefficients)

    def __call__(self, coefficients):
        if isinstance(coefficients, int):
            return self.element(self, [coefficients])
        if not isinstance(coefficients, list):
            raise TypeError(
                f"Polynomials should be constructed from a list of integers, of length at most d = {self.n}"
            )
        return self.element(self, coefficients)

    def __repr__(self):
        return f"Univariate Polynomial Ring in x over Finite Field of size {self.q} with modulus x^{self.n} + 1"


class GenericPolynomial:
    def __init__(self, parent, coefficients):
        self.parent = parent
        self.coeffs = self._parse_coefficients(coefficients)

    def is_zero(self):
        """
        Return if polynomial is zero: f = 0
        """
        return all(c == 0 for c in self.coeffs)

    def is_constant(self):
        """
        Return if polynomial is constant: f = c
        """
        return all(c == 0 for c in self.coeffs[1:])

    def _parse_coefficients(self, coefficients):
        """
        Helper function which right pads with zeros
        to allow polynomial construction as
        f = R([1,1,1])
        """
        l = len(coefficients)
        if l > self.parent.n:
            raise ValueError(
                f"Coefficients describe polynomial of degree greater than maximum degree {self.parent.n}"
            )
        elif l < self.parent.n:
            coefficients = coefficients + [0 for _ in range(self.parent.n - l)]
        return coefficients

    def reduce_coefficients(self):
        """
        Reduce all coefficients modulo q
        """
        self.coeffs = [c % self.parent.q for c in self.coeffs]
        return self

    def _add_mod_q(self, x, y):
        """
        add two coefficients modulo q
        """
        return (x + y) % self.parent.q

    def _sub_mod_q(self, x, y):
        """
        sub two coefficients modulo q
        """
        return (x - y) % self.parent.q

    def _schoolbook_multiplication(self, other):
        """
        Naive implementation of polynomial multiplication
        suitible for all R_q = F_1[X]/(X^n + 1)
        """
        n = self.parent.n
        a = self.coeffs
        b = other.coeffs
        new_coeffs = [0 for _ in range(n)]
        for i in range(n):
            for j in range(0, n - i):
                new_coeffs[i + j] += a[i] * b[j]
        for j in range(1, n):
            for i in range(n - j, n):
                new_coeffs[i + j - n] -= a[i] * b[j]
        return [c % self.parent.q for c in new_coeffs]

    def __neg__(self):
        """
        Returns -f, by negating all coefficients
        """
        neg_coeffs = [(-x % self.parent.q) for x in self.coeffs]
        return self.parent(neg_coeffs)

    def _add_(self, other):
        if isinstance(other, type(self)):
            new_coeffs = [
                self._add_mod_q(x, y)
                for x, y in zip(self.coeffs, other.coeffs)
            ]
        elif isinstance(other, int):
            new_coeffs = self.coeffs.copy()
            new_coeffs[0] = self._add_mod_q(new_coeffs[0], other)
        else:
            raise NotImplementedError(
                "Polynomials can only be added to each other"
            )
        return new_coeffs

    def __add__(self, other):
        new_coeffs = self._add_(other)
        return self.parent(new_coeffs)

    def __radd__(self, other):
        return self.__add__(other)

    def __iadd__(self, other):
        self = self + other
        return self

    def _sub_(self, other):
        if isinstance(other, type(self)):
            new_coeffs = [
                self._sub_mod_q(x, y)
                for x, y in zip(self.coeffs, other.coeffs)
            ]
        elif isinstance(other, int):
            new_coeffs = self.coeffs.copy()
            new_coeffs[0] = self._sub_mod_q(new_coeffs[0], other)
        else:
            raise NotImplementedError(
                "Polynomials can only be subtracted from each other"
            )
        return new_coeffs

    def __sub__(self, other):
        new_coeffs = self._sub_(other)
        return self.parent(new_coeffs)

    def __rsub__(self, other):
        return -self.__sub__(other)

    def __isub__(self, other):
        self = self - other
        return self

    def __mul__(self, other):
        if isinstance(other, type(self)):
            new_coeffs = self._schoolbook_multiplication(other)
        elif isinstance(other, int):
            new_coeffs = [(c * other) % self.parent.q for c in self.coeffs]
        else:
            raise NotImplementedError(
                "Polynomials can only be multiplied by each other, or scaled by integers"
            )
        return self.parent(new_coeffs)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __imul__(self, other):
        self = self * other
        return self

    def __pow__(self, n):
        if not isinstance(n, int):
            raise TypeError(
                "Exponentiation of a polynomial must be done using an integer."
            )

        # Deal with negative scalar multiplication
        if n < 0:
            raise ValueError(
                "Negative powers are not supported for elements of a Polynomial Ring"
            )
        f = self
        g = self.parent(1)
        while n > 0:
            if n % 2 == 1:
                g = g * f
            f = f * f
            n = n // 2
        return g

    def __eq__(self, other):
        if isinstance(other, type(self)):
            return self.coeffs == other.coeffs
        elif isinstance(other, int):
            if (
                self.is_constant()
                and (other % self.parent.q) == self.coeffs[0]
            ):
                return True
        return False

    def __getitem__(self, idx):
        return self.coeffs[idx]

    def __repr__(self):
        if self.is_zero():
            return "0"

        info = []
        for i, c in enumerate(self.coeffs):
            if c != 0:
                if i == 0:
                    info.append(f"{c}")
                elif i == 1:
                    if c == 1:
                        info.append("x")
                    else:
                        info.append(f"{c}*x")
                else:
                    if c == 1:
                        info.append(f"x^{i}")
                    else:
                        info.append(f"{c}*x^{i}")
        return " + ".join(info)

    def __str__(self):
        return self.__repr__()



import os
from ..utilities.utils import xor_bytes
from Crypto.Cipher import AES
from typing import Optional


class AES256_CTR_DRBG:
    def __init__(
        self, seed: Optional[bytes] = None, personalization: bytes = b""
    ):
        """
        DRBG implementation based on AES-256 CTR following the document NIST SP
        800-90A Section 10.2.1

        https://csrc.nist.gov/pubs/sp/800/90/a/r1/final

        Used for deterministic randomness, particularly used for comparing the
        output of Kyber/ML-KEM against known answer tests.

        :param bytes seed: 48 byte seed, if none is supplied a seed is generated
            using ``os.urandom(48)``.
        :param bytes personalization: optional bytes, of length at most 48 used
            during instantiation of the DRBG
        """
        self.seed_length = 48
        self.reseed_interval = 2**48
        self.key = bytes([0]) * 32
        self.V = bytes([0]) * 16
        self.entropy_input = self.__check_entropy_input(seed)

        seed_material = self.__instantiate(personalization=personalization)
        self.__ctr_drbg_update(seed_material)
        self.reseed_ctr = 1

    def __check_entropy_input(self, entropy_input: bytes) -> bytes:
        """
        If no entropy given, us os.urandom, else
        check that the input is of the right length.
        """
        if entropy_input is None:
            return os.urandom(self.seed_length)
        elif len(entropy_input) != self.seed_length:
            raise ValueError(
                f"The entropy input must be of length: {self.seed_length}. "
                f"Input has length {len(entropy_input)}"
            )
        return entropy_input

    def __instantiate(self, personalization: bytes = b"") -> bytes:
        """
        Combine the input seed and optional personalisation
        string into the seed material for the DRBG

        Section 10.2.1.3.1, Page 52 (CTR_DRBG_Instantiate_algorithm)
        """
        if len(personalization) > self.seed_length:
            raise ValueError(
                f"The Personalization String must be at most length: "
                f"{self.seed_length}. Input has length {len(personalization)}"
            )
        # Ensure personalization has exactly seed_length bytes
        personalization += bytes([0]) * (
            self.seed_length - len(personalization)
        )
        # debugging
        assert len(personalization) == self.seed_length
        return xor_bytes(self.entropy_input, personalization)

    def __increment_counter(self) -> None:
        """
        Increment the internal counter of the DRBG
        """
        int_V = int.from_bytes(self.V, "big")
        new_V = (int_V + 1) % 2**128
        self.V = new_V.to_bytes(16, byteorder="big")

    def __ctr_drbg_update(self, provided_data: bytes) -> None:
        """
        Updates the internal state of the CTR_DRBG using the
        provided_data

        Section 10.2.1.2, Page 51 (CTR_DRBG_Update)
        """
        tmp = b""
        cipher = AES.new(self.key, AES.MODE_ECB)

        # Collect bytes from AES ECB
        while len(tmp) != self.seed_length:
            self.__increment_counter()
            tmp += cipher.encrypt(self.V)

        # Take the first 48 bytes
        tmp = tmp[: self.seed_length]
        tmp = xor_bytes(tmp, provided_data)

        # Set the new values of key and V
        self.key = tmp[:32]
        self.V = tmp[32:]

    def random_bytes(
        self, num_bytes: int, additional: Optional[bytes] = None
    ) -> bytes:
        """
        Generate pseudorandom bytes without a generating function

        Section 10.2.1.5.1, Page 56 (CTR_DRBG_Generate_algorithm)

        :param int num_bytes: the number of random bytes requested
        :param bytes additional: optional bytes to be mixed into the generation
        :return: pseudorandom bytes extracted from the DRBG of length ``num_bytes``.
        :rtype: bytes
        """
        # We don't cover this in coverage as we would need to run the counter 2^48 times
        if self.reseed_ctr >= self.reseed_interval:  # pragma: no cover
            raise Warning("The DRBG has been exhausted! Reseed!")

        # Set the optional additional information
        if additional is None:
            additional = bytes([0]) * self.seed_length
        else:
            if len(additional) > self.seed_length:
                raise ValueError(
                    f"The additional input must be of length at most: "
                    f"{self.seed_length}. Input has length {len(additional)}"
                )
            additional += bytes([0]) * (self.seed_length - len(additional))
            self.__ctr_drbg_update(additional)

        # Collect bytes!
        tmp = b""
        cipher = AES.new(self.key, AES.MODE_ECB)
        while len(tmp) < num_bytes:
            self.__increment_counter()
            tmp += cipher.encrypt(self.V)

        # Collect only the requested number of bits
        output_bytes = tmp[:num_bytes]
        self.__ctr_drbg_update(additional)
        self.reseed_ctr += 1
        return output_bytes



from ..utilities.utils import bit_count
from .polynomials_generic import GenericPolynomialRing, GenericPolynomial


class PolynomialRing(GenericPolynomialRing):
    """
    Initialise the polynomial ring:

        R = GF(3329) / (X^256 + 1)
    """

    def __init__(self):
        self.q = 3329
        self.n = 256
        self.element = Polynomial
        self.element_ntt = PolynomialNTT

        root_of_unity = 17
        self.ntt_zetas = [
            pow(root_of_unity, self._br(i, 7), 3329) for i in range(128)
        ]
        self.ntt_f = pow(128, -1, 3329)

    @staticmethod
    def _br(i, k):
        """
        bit reversal of an unsigned k-bit integer
        """
        bin_i = bin(i & (2**k - 1))[2:].zfill(k)
        return int(bin_i[::-1], 2)

    def ntt_sample(self, input_bytes):
        """
        Algorithm 1 (Parse)
        https://pq-crystals.org/kyber/data/kyber-specification-round3-20210804.pdf

        Algorithm 6 (Sample NTT)

        Parse: B^* -> R
        """
        i, j = 0, 0
        coefficients = [0 for _ in range(self.n)]
        while j < self.n:
            d1 = input_bytes[i] + 256 * (input_bytes[i + 1] % 16)
            d2 = (input_bytes[i + 1] // 16) + 16 * input_bytes[i + 2]

            if d1 < 3329:
                coefficients[j] = d1
                j = j + 1

            if d2 < 3329 and j < self.n:
                coefficients[j] = d2
                j = j + 1

            i = i + 3
        return self(coefficients, is_ntt=True)

    def cbd(self, input_bytes, eta, is_ntt=False):
        """
        Algorithm 2 (Centered Binomial Distribution)
        https://pq-crystals.org/kyber/data/kyber-specification-round3-20210804.pdf

        Algorithm 6 (Sample Poly CBD)

        Expects a byte array of length (eta * deg / 4)
        For Kyber, this is 64 eta.
        """
        assert 64 * eta == len(input_bytes)
        coefficients = [0 for _ in range(256)]
        b_int = int.from_bytes(input_bytes, "little")
        mask = (1 << eta) - 1
        mask2 = (1 << 2 * eta) - 1
        for i in range(256):
            x = b_int & mask2
            a = bit_count(x & mask)
            b = bit_count((x >> eta) & mask)
            b_int >>= 2 * eta
            coefficients[i] = (a - b) % 3329
        return self(coefficients, is_ntt=is_ntt)

    def decode(self, input_bytes, d, is_ntt=False):
        """
        Decode (Algorithm 3)

        decode: B^32l -> R_q
        """
        # Ensure the value d is set correctly
        if 256 * d != len(input_bytes) * 8:
            raise ValueError(
                f"input bytes must be a multiple of (polynomial degree) / 8, {256*d = }, {len(input_bytes)*8 = }"
            )

        # Set the modulus
        if d == 12:
            m = 3329
        else:
            m = 1 << d

        coeffs = [0 for _ in range(256)]
        b_int = int.from_bytes(input_bytes, "little")
        mask = (1 << d) - 1
        for i in range(256):
            coeffs[i] = (b_int & mask) % m
            b_int >>= d

        return self(coeffs, is_ntt=is_ntt)

    def __call__(self, coefficients, is_ntt=False):
        if not is_ntt:
            element = self.element
        else:
            element = self.element_ntt

        if isinstance(coefficients, int):
            return element(self, [coefficients])
        if not isinstance(coefficients, list):
            raise TypeError(
                f"Polynomials should be constructed from a list of integers, of length at most n = {256}"
            )
        return element(self, coefficients)


class Polynomial(GenericPolynomial):
    def __init__(self, parent, coefficients):
        self.parent = parent
        self.coeffs = self._parse_coefficients(coefficients)

    def encode(self, d):
        """
        Encode (Inverse of Algorithm 3)
        """
        t = 0
        for i in range(255):
            t |= self.coeffs[256 - i - 1]
            t <<= d
        t |= self.coeffs[0]
        return t.to_bytes(32 * d, "little")

    def _compress_ele(self, x, d):
        """
        Compute round((2^d / q) * x) % 2^d
        """
        t = 1 << d
        y = (t * x + 1664) // 3329  # 1664 = 3329 // 2
        return y % t

    def _decompress_ele(self, x, d):
        """
        Compute round((q / 2^d) * x)
        """
        t = 1 << (d - 1)
        y = (3329 * x + t) >> d
        return y

    def compress(self, d):
        """
        Compress the polynomial by compressing each coefficient

        NOTE: This is lossy compression
        """
        self.coeffs = [self._compress_ele(c, d) for c in self.coeffs]
        return self

    def decompress(self, d):
        """
        Decompress the polynomial by decompressing each coefficient

        NOTE: This as compression is lossy, we have
        x' = decompress(compress(x)), which x' != x, but is
        close in magnitude.
        """
        self.coeffs = [self._decompress_ele(c, d) for c in self.coeffs]
        return self

    def to_ntt(self):
        """
        Convert a polynomial to number-theoretic transform (NTT) form.
        The input is in standard order, the output is in bit-reversed order.
        """
        k, l = 1, 128
        coeffs = self.coeffs
        zetas = self.parent.ntt_zetas
        while l >= 2:
            start = 0
            while start < 256:
                zeta = zetas[k]
                k = k + 1
                j = start
                for j in range(start, start + l):
                    t = zeta * coeffs[j + l]
                    coeffs[j + l] = coeffs[j] - t
                    coeffs[j] = coeffs[j] + t
                start = l + (j + 1)
            l = l >> 1

        for j in range(256):
            coeffs[j] = coeffs[j] % 3329

        return self.parent(coeffs, is_ntt=True)

    def from_ntt(self):
        """
        Not supported, raises a ``TypeError``
        """
        raise TypeError(f"Polynomial not in the NTT domain: {type(self) = }")


class PolynomialNTT(Polynomial):
    def __init__(self, parent, coefficients):
        self.parent = parent
        self.coeffs = self._parse_coefficients(coefficients)

    def to_ntt(self):
        """
        Not supported, raises a ``TypeError``
        """
        raise TypeError(
            f"Polynomial is already in the NTT domain: {type(self) = }"
        )

    def from_ntt(self):
        """
        Convert a polynomial from number-theoretic transform (NTT) form in place
        The input is in bit-reversed order, the output is in standard order.
        """
        l, l_upper = 2, 128
        k = l_upper - 1
        coeffs = self.coeffs
        zetas = self.parent.ntt_zetas
        while l <= 128:
            start = 0
            while start < 256:
                zeta = zetas[k]
                k = k - 1
                j = start
                for j in range(start, start + l):
                    t = coeffs[j]
                    coeffs[j] = t + coeffs[j + l]
                    coeffs[j + l] = coeffs[j + l] - t
                    coeffs[j + l] = zeta * coeffs[j + l]
                start = j + l + 1
            l = l << 1

        f = self.parent.ntt_f
        for j in range(256):
            coeffs[j] = (coeffs[j] * f) % 3329

        return self.parent(coeffs, is_ntt=False)

    @staticmethod
    def _ntt_base_multiplication(a0, a1, b0, b1, zeta):
        """
        Base case for ntt multiplication
        """
        r0 = (a0 * b0 + zeta * a1 * b1) % 3329
        r1 = (a1 * b0 + a0 * b1) % 3329
        return r0, r1

    def _ntt_coefficient_multiplication(self, f_coeffs, g_coeffs):
        """
        Given the coefficients of two polynomials compute the coefficients of
        their product
        """
        new_coeffs = []
        zetas = self.parent.ntt_zetas
        for i in range(64):
            r0, r1 = self._ntt_base_multiplication(
                f_coeffs[4 * i + 0],
                f_coeffs[4 * i + 1],
                g_coeffs[4 * i + 0],
                g_coeffs[4 * i + 1],
                zetas[64 + i],
            )
            r2, r3 = self._ntt_base_multiplication(
                f_coeffs[4 * i + 2],
                f_coeffs[4 * i + 3],
                g_coeffs[4 * i + 2],
                g_coeffs[4 * i + 3],
                -zetas[64 + i],
            )
            new_coeffs += [r0, r1, r2, r3]
        return new_coeffs

    def _ntt_multiplication(self, other):
        """
        Number Theoretic Transform multiplication.
        """
        new_coeffs = self._ntt_coefficient_multiplication(
            self.coeffs, other.coeffs
        )
        return new_coeffs

    def __add__(self, other):
        new_coeffs = self._add_(other)
        return self.parent(new_coeffs, is_ntt=True)

    def __sub__(self, other):
        new_coeffs = self._sub_(other)
        return self.parent(new_coeffs, is_ntt=True)

    def __mul__(self, other):
        if isinstance(other, type(self)):
            new_coeffs = self._ntt_multiplication(other)
        elif isinstance(other, int):
            new_coeffs = [(c * other) % 3329 for c in self.coeffs]
        else:
            raise NotImplementedError(
                f"Polynomials can only be multiplied by each other, or scaled by integers, {type(other) = }, {type(self) = }"
            )
        return self.parent(new_coeffs, is_ntt=True)



import sys

# int.bit_count() was only made available in 3.10
if sys.version_info >= (3, 10):

    def bit_count(x: int) -> int:
        """
        Count the number of bits in x
        """
        return x.bit_count()

else:

    def bit_count(x: int) -> int:
        """
        Count the number of bits in x
        """
        return bin(x).count("1")


def xor_bytes(a, b):
    """
    XOR two byte arrays, assume that they are
    of the same length
    """
    assert len(a) == len(b)
    return bytes(a ^ b for a, b in zip(a, b))


def select_bytes(a, b, cond):
    """
    Select between the bytes a or b depending
    on whether cond is False or True
    """
    assert len(a) == len(b)
    out = [0] * len(a)
    cw = -cond % 256
    for i in range(len(a)):
        out[i] = a[i] ^ (cw & (a[i] ^ b[i]))
    return bytes(out)



"""
Import and export of keys to PKCS #8 and SPKI format.

This module allows for importing private keys (decapsulation keys) from
PKCS #8 files, exporting them to PKCS #8 files, and both import and export
of the public keys (encapsulation keys) from the Subject Public Key Info
format used in X.509 certificates and in bare public keys.
"""

try:
    from ecdsa import der
except ImportError:
    raise ImportError("PKCS functionality requires the ecdsa library")

from .default_parameters import ML_KEM_512, ML_KEM_768, ML_KEM_1024


OIDS = {
    ML_KEM_512.oid: ML_KEM_512,
    ML_KEM_768.oid: ML_KEM_768,
    ML_KEM_1024.oid: ML_KEM_1024,
}


def ek_to_der(kem, ek):
    """
    Convert an encapsulation key to a SPKI DER structure.

    :param kem: an ``ML_KEM`` object instance
    :param bytes ek: encapsulation key
    :rtype: bytes
    """
    if not kem.oid:
        raise ValueError("Only KEMs with specified OIDs can be encoded")

    if len(ek) != kem._ek_size():
        raise ValueError("Provided key size doesn't match the provided kem")

    enc = der.encode_sequence(
        der.encode_sequence(
            der.encode_oid(*kem.oid),
        ),
        der.encode_bitstring(ek, 0),
    )

    return enc


def ek_to_pem(kem, ek):
    """
    Convert an encapsulation key to a SPKI PEM structure.

    :param kem: an ``ML_KEM`` object instance
    :param bytes ek: encapsulation key
    :rtype: str
    """
    der_enc = ek_to_der(kem, ek)

    pem_enc = der.topem(der_enc, "PUBLIC KEY")

    return pem_enc


def ek_from_der(enc_key):
    """
    Extract an encapsulation key from DER encoding.

    :param bytes enc_key: SPKI DER encoding of a key
    :rtype: tuple(ML_KEM, bytes)
    """
    s1, empty = der.remove_sequence(enc_key)
    if empty:
        raise der.UnexpectedDER("Trailing junk after DER public key")

    alg_id, rem = der.remove_sequence(s1)

    alg_id, rest = der.remove_object(alg_id)
    if alg_id not in OIDS:
        raise der.UnexpectedDER(f"Not recognised algoritm OID: {alg_id}")
    if rest:
        raise der.UnexpectedDER("Parameters specified for ML-KEM OID")

    kem = OIDS[alg_id]

    key, empty = der.remove_bitstring(rem, 0)
    if empty:
        raise der.UnexpectedDER("Trailing junk after public key bitsting")

    if len(key) != kem._ek_size():
        raise der.UnexpectedDER("Wrong key size for the OID in structure")

    return kem, key


def ek_from_pem(enc_key):
    """
    Extract an encapsulation key from PEM encoding.

    :param str enc_key: SPKI PEM encoding of a key
    :rtype: tuple(ML_KEM, bytes)
    """
    der_key = der.unpem(enc_key)
    return ek_from_der(der_key)


def dk_to_der(kem, dk=None, seed=None, form=None):
    """
    Convert the decapsulation key to a PKCS #8 DER structure.

    ``ek``, ``seed``, or both need to be provided.
    If ``form`` is not specified (is set to ``None``), the format that
    preserves maximum amount of information will be used.

    Proposed in
    https://mailarchive.ietf.org/arch/msg/spasm/50v8oLi5XObC7AIL4DH337_Anos/
    for the draft-ietf-lamps-kyber-certificates

    :param kem: an ``ML_KEM`` object instance
    :param bytes dk: decapsulation key
    :param bytes seed: seed to generate ML-KEM keys
    :param str form: What format to write the key in, options are:
        None - for automatic selection based on ``dk`` and ``seed``,
        ``seed`` for writing seed only, ``expanded`` for writing
        the expanded key only, ``both`` for writing both key and expanded key.
    :rtype: bytes
    """
    if form not in ("seed", "expanded", "both", None):
        raise ValueError(
            f"Invalid form specified: {form}. "
            "Only 'seed', 'expanded', 'both', and None are allowed"
        )
    if not dk and not seed:
        raise ValueError("dk or seed must be provided")

    if dk and len(dk) != kem._dk_size():
        raise ValueError("Invalid decapsulation key size for the provided KEM")

    if seed and len(seed) != 64:
        raise ValueError("Invalid seed size")

    if form in ("both", "seed") and not seed:
        raise ValueError(f'Format "{form}" requires specifing seed')

    if form is None:
        if dk and seed:
            form = "both"
        elif dk:
            form = "expanded"
        else:
            assert seed
            form = "seed"

    if form in ("both", "seed") and not dk:
        _, dk = kem.key_derive(seed)

    if form == "seed":
        enc_key = der.encode_implicit(0, seed)
    elif form == "expanded":
        enc_key = der.encode_octet_string(dk)
    else:
        assert form == "both"
        enc_key = der.encode_sequence(
            der.encode_octet_string(seed), der.encode_octet_string(dk)
        )

    encoded_pkcs8 = der.encode_sequence(
        der.encode_integer(0),
        der.encode_sequence(der.encode_oid(*kem.oid)),
        der.encode_octet_string(enc_key),
    )

    return encoded_pkcs8


def dk_to_pem(kem, dk=None, seed=None, form=None):
    """
    Convert the decapsulation key to a PKCS #8 DER structure.

    ``ek``, ``seed``, or both need to be provided.
    If ``form`` is not specified (is set to ``None``), the format that
    preserves maximum amount of information will be used.

    Proposed in
    https://mailarchive.ietf.org/arch/msg/spasm/50v8oLi5XObC7AIL4DH337_Anos/
    for the draft-ietf-lamps-kyber-certificates

    :param kem: an ``ML_KEM`` object instance
    :param bytes dk: decapsulation key
    :param bytes seed: seed to generate ML-KEM keys
    :param str form: What format to write the key in, options are:
        None - for automatic selection based on ``dk`` and ``seed``,
        ``seed`` for writing seed only, ``expanded`` for writing
        the expanded key only, ``both`` for writing both key and expanded key.
    :rtype: bytes
    """
    der_enc = dk_to_der(kem, dk, seed, form)

    pem_enc = der.topem(der_enc, "PRIVATE KEY")

    return pem_enc


def dk_from_der(enc_key):
    """
    Extract encapsulation and decapsulation keys from PKCS #8 DER encoding.

    :param bytes enc_key: PKCS #8 DER encoding of the key
    :return: the first element returned is the ``ML_KEM`` object instance,
        second element is the decapsulation key, third is the seed (if present),
        and fourth is the encapsulation key.
    :rtype: tuple(ML_KEM, bytes, bytes, bytes)
    """
    s1, empty = der.remove_sequence(enc_key)
    if empty:
        raise der.UnexpectedDER("Trailing junk after private key structure")

    ver, rest = der.remove_integer(s1)

    if ver != 0:
        raise der.UnexpectedDER(f"Unsupported version: {ver}")

    alg_id, rest = der.remove_sequence(rest)

    alg_id, empty = der.remove_object(alg_id)
    if alg_id not in OIDS:
        raise der.UnexpectedDER(f"Not recognised algorithm OID: {alg_id}")
    if empty:
        raise der.UnexpectedDER("Junk after algorithm OID")

    kem = OIDS[alg_id]

    priv_key, _ = der.remove_octet_string(rest)
    # "rest" here can be either parameters of public key: we ignore those

    seed = None
    expanded = None
    ek = None

    if der.str_idx_as_int(priv_key, 0) == 0x04:
        # we have OCTET STRING: expanded only format
        expanded, empty = der.remove_octet_string(priv_key)
        if empty:
            raise der.UnexpectedDER("Junk after expandedKey")
    elif der.is_sequence(priv_key):
        both, empty = der.remove_sequence(priv_key)
        if empty:
            raise der.UnexpectedDER("Junk after both encoding")
        seed, key_val = der.remove_octet_string(both)
        expanded, empty = der.remove_octet_string(key_val)
        if empty:
            raise der.UnexpectedDER("Junk after 'expandedKey' in 'both' value")
    else:
        tag, seed, empty = der.remove_implicit(priv_key)
        if tag != 0:
            raise der.UnexpectedDER("Unexpected tag in private key encoding")
        if empty:
            raise der.UnexpectedDER("Junk after seed encoding")

    if expanded and len(expanded) != kem._dk_size():
        raise der.UnexpectedDER("Invalid expanded key size in encoding")

    if not expanded:
        ek, expanded = kem.key_derive(seed)

    if seed and len(seed) != 64:
        raise der.UnexpectedDER("Invalid length of seed in encoding")

    if not ek:
        ek = expanded[384 * kem.k : 768 * kem.k + 32]

    return kem, expanded, seed, ek


def dk_from_pem(enc_key):
    """
    Extract encapsulation and decapsulation keys from PKCS #8 PEM encoding.

    :param str enc_key: PKCS #8 PEM encoding of the key
    :return: the first element returned is the ``ML_KEM`` object instance,
        second element is the decapsulation key, third is the seed (if present),
        and fourth is the encapsulation key.
    :rtype: tuple(ML_KEM, bytes, bytes, bytes)
    """
    der_key = der.unpem(enc_key)
    return dk_from_der(der_key)
